==> ump-0.8.3.2-src.tar.bz2: ump-0.8.3.2/src/complex.cpp <==

/*
 *  Complex - Implements a complex number
 *  Copyright (c) 2003-2005 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See complex.h
 */

/*
News
----

v1  2005-09-25 - 2005-10-10
--

	Extracted class Complex from the math2.* system
	ump_float -> floatx
	operator < & > doesn't allow imaginary part to be non-zero anymore

*/

#include "complex.h"


void Complex::append_to_string( std::string &str, const Format &fmt ) const throw(error_obj)
{
	std::string tmp_str;

	try
	{
		if( floatx(imaginary) == 0.0 )
			real.append_to_string( str, fmt );
		else
		{
			if( floatx(real) != 0.0 )
			{
				real.append_to_string( str, fmt );

				if( floatx(imaginary) < 0.0 )
				{
					if( floatx(imaginary) == -1.0 )
					{
						tmp_str = " - i";
						str.append( tmp_str );
					}
					else
					{
						
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = " - (";
						else
							tmp_str = " - ";
						str.append( tmp_str );
						(-imaginary).append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = ")i";
						else
							tmp_str = "i";
						str.append( tmp_str );
					}
				}
				else
				{
					if( floatx(imaginary) == 1.0 )
					{
						tmp_str = " + i";
						str.append( tmp_str );
					}
					else
					{
						
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str =  " + (";
						else
							tmp_str = " + ";
						str.append( tmp_str );
						imaginary.append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = ")i";
						else
							tmp_str = "i";
						str.append( tmp_str );
					}
				}
			}
			else	
			{
				if( floatx(imaginary) < 0.0 )
				{
					if( floatx(imaginary) == -1.0 )
					{
						tmp_str = "-i";
						str.append( tmp_str );
					}
					else
					{
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = "-(";
						else
							tmp_str = "-";
						str.append( tmp_str );
						(-imaginary).append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = ")i";
						else
							tmp_str = "i";
						str.append( tmp_str );
					}
				}
				else
				{
					if( floatx(imaginary) == 1.0 )
					{
						tmp_str = "i";
						str.append( tmp_str );
					}
					else
					{
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
						{
							tmp_str = "(";
							str.append( tmp_str );
						}
						imaginary.append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							tmp_str = ")i";
						else
							tmp_str = "i";
						str.append( tmp_str );
					}
				}
			}
		}
	}
	catch(error_obj error) {  throw error;                 }
	catch(...)             {  Throw_Couldnt_get_memory();  }
}

Complex Complex::operator*(const Complex &var) const
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return Complex( real*var.real );
	else
	{
		Complex res( real, real );
		Real tmp( imaginary );
		res.real.mul( var.real );
		tmp.mul( var.imaginary );
		res.real.sub( tmp );

		res.imaginary.mul( var.imaginary );
		tmp = imaginary;
		tmp.mul( var.real );
		res.imaginary.add( tmp );

		return res;
	}
}

Complex Complex::operator/(const Complex &var) const throw(error_obj)
{
	if( floatx(var.real) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		Throw_Divide_by_zero();

	if( floatx(var.imaginary) == 0.0 )
		return Complex( real/var.real, imaginary/var.real);
	else
	{
		Real tmp = var.real*var.real + var.imaginary*var.imaginary;
		return Complex( (real*var.real + imaginary*var.imaginary) / tmp ,
		       (imaginary*var.real - real*var.imaginary) / tmp );
	}
}

Complex operator/(const Real &val,const Complex &com)
{
	if( floatx(com.real) == 0.0  &&  floatx(com.imaginary) == 0.0 )
		Throw_Divide_by_zero();

	if(floatx(com.imaginary) == 0.0)
		return Complex( val/com.real );
	else
		return Complex(val*com.real / (com.real*com.real + com.imaginary*com.imaginary),
		       -val*com.imaginary/ (com.real*com.real + com.imaginary*com.imaginary));
}

bool Complex::operator<(const Complex &var) const throw(error_obj)
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return (real < var.real);
	else
		Throw_Domain_error( "imaginary must be zero" );
}

bool Complex::operator>(const Complex &var) const throw(error_obj)
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return (real > var.real);
	else
		Throw_Domain_error( "imaginary must be zero" );
}

Complex sqroot(const Complex &var) throw(error_obj)
{
	if( floatx(var.imaginary) != 0.0 )
		return power( var, Complex( 0.5 ) );

	if( floatx(var.real) >= 0.0 )
		return Complex( sqrtx(floatx(var.real)) );
	else
		return Complex( Real(0), sqrtx(-floatx(var.real)) );
}

bool complex_almost_equal(const Complex &left,const Complex &right)
{
	Complex l,r;

	r = right;
	l = left;

	if( floatx(l.real) <= 1e-10  &&  floatx(l.real) >= -1e-10 )
		l.real = Real(1e-10);
	if( floatx(l.imaginary) <= 1e-10  && floatx(l.imaginary) >= -1e-10 )
		l.imaginary = Real(1e-10);
	if( floatx(r.real) <= 1e-10  &&  floatx(r.real) >= -1e-10 )
		r.real = Real(1e-10);
	if( floatx(r.imaginary) <= 1e-10  &&  floatx(r.imaginary) >= -1e-10 )
		r.imaginary = Real(1e-10);

	if( (floatx(l.real) * floatx(r.real)) < 0.0 ) // not same sign
		return false;
	else if( (floatx(l.imaginary) * floatx(r.imaginary)) < 0.0 )
		return false;
	else if( fabsx(log10x(floatx(l.real)) - log10x(floatx(r.real))) > 1e-10 )
		return false;
	else if( fabsx(log10x(floatx(l.imaginary)) - log10x(floatx(r.imaginary))) > 1e-10 )
		return false;

	return true;
}

Complex power(const Complex &left,const Complex &right) throw(error_obj)
{
	Complex tmp;

	tmp.imaginary = Real(0);

	if( floatx(right.imaginary) != 0.0 )
	{
		// let's use the fact that   a^b = ( e^(ln a) )^b = e^(b*ln a)
		tmp = expc( right * logc(left) );
	}
	else
	{
		if( floatx(left.imaginary) != 0.0 ) // we need to use de Moivres formula
		{
de_moivres:
			Complex tmp_abs,tmp_arg;

			tmp_abs = abs(left);
			tmp_arg = arg(left);
			tmp = powx( floatx(tmp_abs.real), floatx(right.real) ) * 
			           (Complex( cosx(floatx(right.real)*floatx(tmp_arg.real)),
			           sinx(floatx(right.real)*floatx(tmp_arg.real)) ) );
			// tmp = r^n * ( cos(n*phi) + i*sin(n*phi) )
		}
		else
		{
			if( left.real.isExact()  &&  right.real.isInteger() )
			{
				Integer nr;
				right.real.get_top( &nr );
				tmp.real = power( left.real, nr );
			}
			else if( floatx(left.real) < 0.0 )
			{
				if(right.real == ipart(right.real)) // checks if right.real == N
					tmp.real = Real( powx(floatx(left.real), floatx(right.real)) );
				else if((1.0/floatx(right.real)) == (truncx(1.0/floatx(right.real)))) // checks if right.real == 1/N
				{
					if(((1.0/floatx(right.real))/4.0) == (truncx((1.0/floatx(right.real))/4.0))) // N is dividable with 4
						goto de_moivres;
					else if(((1.0/floatx(right.real))/2.0) == (truncx((1.0/floatx(right.real))/2.0))) // N is dividable with 2
					{
						tmp.real = Real(0);
						tmp.imaginary = Real( powx(-floatx(left.real),floatx(right.real)) );
					}
					else
						tmp.real = Real( -powx(-floatx(left.real),floatx(right.real)) );
				}
				else
					goto de_moivres;
			}
			else
				tmp.real = Real( powx(floatx(left.real),floatx(right.real)) );
		}
	}
	return tmp;
}

Complex expc(const Complex &val)
{
	Complex tmp;

	tmp.real = expx( floatx(val.real) );
	tmp.imaginary = tmp.real * Real( sinx( floatx(val.imaginary) ) );
	tmp.real = tmp.real * Real( cosx( floatx(val.imaginary) ) );

	return tmp;
}

Complex logc(const Complex &val)
{
//ln(z) = ln( abs(z) ) + i*arg(z)
	return Complex( logx( floatx( abs( val ).real ) ), arg( val ).real );
}

Complex log2c(const Complex &val)
{
//log2(z) = ln(z)/ln(2)
	return logc( val ) / Complex( logx( 2.0 ) );
}

Complex log10c(const Complex &val)
{
//log10(z) = ln(z)/ln(10)
	return logc( val ) / Complex( logx( 10.0 ) );
}

Complex sinc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( sinx( floatx(val.real) ) );
	else
	{
// sin(z) = ( e^(iz) - e^(-iz) ) / (2i)
		return ( expc(Complex( 0.0, 1.0 ) * val) - expc( Complex( 0.0, -1.0 ) * val ) ) /
		         Complex( 0.0, 2.0 );
	}
}

Complex cosc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( cosx( floatx(val.real) ) );
	else
	{
// cos(z) = ( e^(iz) + e^(-iz) ) / 2
		return ( expc(Complex( 0.0, 1.0 ) * val) + expc( Complex( 0.0, -1.0 ) * val ) ) /
		         Complex( 2.0 );
	}
}

Complex tanc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( tanx( floatx(val.real) ) );
	else
	{
// tan(z) = ( e^(2iz) -1 ) / ( i(e^(2iz) +1) )
		Complex tmp = expc( Complex( 0.0, 2.0 ) * val );
		return (tmp - Complex( 1.0 )) / ( Complex( 0.0, 1.0 ) * (tmp + Complex( 1.0 )) );
	}
}

Complex asinc(const Complex &val)
{
	floatx tmp = floatx(val.real);
	if( floatx(val.imaginary) == 0.0  &&  tmp <= 1.0  &&  tmp >= -1.0 )
		return Complex( cosx( tmp ) );
	else
	{
// asin(z) = -i*ln( iz + sqrt(1-z^2) )
		return Complex( 0.0, -1.0 ) * logc( Complex( 0.0, 1.0 ) * val +
		       sqroot( Complex( 1.0 ) - val*val ) );
	}
}

Complex acosc(const Complex &val)
{
	floatx tmp = floatx(val.real);
	if( floatx(val.imaginary) == 0.0  &&  tmp <= 1.0  &&  tmp >= -1.0 )
		return Complex( cosx( tmp ) );
	else
	{
// acos(z) = pi/2 + i*ln( iz + sqrt(1-z^2) )
		return Complex( PI/2.0 ) + Complex( 0.0, 1.0 ) * logc( Complex( 0.0, 1.0 ) * val + 
		       sqroot( Complex( 1.0 ) - val*val ) );
	}
}

Complex atanc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( atanx( floatx(val.real) ) );
	else
	{
		// atan(z) = i/2 * ( ln(1-iz) - ln(1+iz) )
		return Complex( 0.0, 0.5 ) * ( logc( Complex( 1.0 ) - Complex( 0.0, 1.0 ) * val ) -
		         logc( Complex( 1.0 ) + Complex( 0.0, 1.0 ) * val ) );
	}
}

Complex sinhc(const Complex &val)
{
//sinh(z) = (exp(z) - exp(-z))/2
	return ( expc( val ) - expc( -val ) ) / Complex( 2.0 );
}

Complex coshc(const Complex &val)
{
//cosh(z) = (exp(z) + exp(-z))/2.
	return ( expc( val ) + expc( -val ) ) / Complex( 2.0 );
}

Complex tanhc(const Complex &val)
{
//tanh(z) = sinh(z)/cosh(z)
	return sinhc( val ) / coshc( val );  
}

Complex asinhc(const Complex &val)
{
//asinh(z) = ln(z + sqrt(z^2+1))
	return logc( val + sqroot( val*val + Complex(1.0) ) );
}

Complex acoshc(const Complex &val)
{
//acosh(z)=ln(z + sqrt(z^2-1))
	return logc( val + sqroot( val*val - Complex(1.0) ) );
}

Complex atanhc(const Complex &val)
{
//atanh(z)=ln( (1+z)/(1-z) ) / 2
	return logc( (Complex(1.0)+val)/(Complex(1.0)-val) ) / Complex(2.0);
}

==> ump-0.8.4-src.tar.bz2: ump-0.8.4/src/complex.cpp <==

/*
 *  Complex - Implements a complex number
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See complex.h
 */

/*
News
----

v2  2006-05-07  &  2006-05-21
--

	Fixed so that abs returns exact results when the argument is exact with imaginary set to zero
	Small changes in operator<= and operator>=
	Small clean up in Complex::append_to_string

v1  2005-09-25 - 2005-10-10
--

	Extracted class Complex from the math2.* system
	ump_float -> floatx
	operator < & > doesn't allow imaginary part to be non-zero anymore

*/

#include "complex.h"


void Complex::append_to_string( std::string &str, const Format &fmt ) const throw(error_obj)
{
	try
	{
		if( floatx(imaginary) == 0.0 )
			real.append_to_string( str, fmt );
		else
		{
			if( floatx(real) != 0.0 )
			{
				real.append_to_string( str, fmt );

				if( floatx(imaginary) < 0.0 )
				{
					if( floatx(imaginary) == -1.0 )
						str.append( " - i" );
					else
					{
						
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( " - (" );
						else
							str.append( " - " );
						(-imaginary).append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( ")i" );
						else
							str.append( "i" );
					}
				}
				else
				{
					if( floatx(imaginary) == 1.0 )
						str.append( " + i" );
					else
					{
						
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( " + (" );
						else
							str.append( " + " );
						imaginary.append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( ")i" );
						else
							str.append( "i" );
					}
				}
			}
			else	
			{
				if( floatx(imaginary) < 0.0 )
				{
					if( floatx(imaginary) == -1.0 )
						str.append( "-i" );
					else
					{
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( "-(" );
						else
							str.append( "-" );
						(-imaginary).append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( ")i" );
						else
							str.append( "i" );
					}
				}
				else
				{
					if( floatx(imaginary) == 1.0 )
						str.append( "i" );
					else
					{
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( "(" );
						imaginary.append_to_string( str, fmt );
						if( !imaginary.isInteger()  &&  imaginary.isExact() )
							str.append( ")i" );
						else
							str.append( "i" );
					}
				}
			}
		}
	}
	catch(error_obj error) {  throw error;                 }
	catch(...)             {  Throw_Couldnt_get_memory();  }
}

Complex Complex::operator*(const Complex &var) const
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return Complex( real*var.real );
	else
	{
		Complex res( real, real );
		Real tmp( imaginary );
		res.real.mul( var.real );
		tmp.mul( var.imaginary );
		res.real.sub( tmp );

		res.imaginary.mul( var.imaginary );
		tmp = imaginary;
		tmp.mul( var.real );
		res.imaginary.add( tmp );

		return res;
	}
}

Complex Complex::operator/(const Complex &var) const throw(error_obj)
{
	if( floatx(var.real) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		Throw_Divide_by_zero();

	if( floatx(var.imaginary) == 0.0 )
		return Complex( real/var.real, imaginary/var.real);
	else
	{
		Real tmp = var.real*var.real + var.imaginary*var.imaginary;
		return Complex( (real*var.real + imaginary*var.imaginary) / tmp ,
		       (imaginary*var.real - real*var.imaginary) / tmp );
	}
}

Complex operator/(const Real &val,const Complex &com)
{
	if( floatx(com.real) == 0.0  &&  floatx(com.imaginary) == 0.0 )
		Throw_Divide_by_zero();

	if(floatx(com.imaginary) == 0.0)
		return Complex( val/com.real );
	else
		return Complex(val*com.real / (com.real*com.real + com.imaginary*com.imaginary),
		       -val*com.imaginary/ (com.real*com.real + com.imaginary*com.imaginary));
}

bool Complex::operator<(const Complex &var) const throw(error_obj)
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return (real < var.real);
	else
		Throw_Domain_error( "imaginary must be zero" );
}

bool Complex::operator>(const Complex &var) const throw(error_obj)
{
	if( floatx(imaginary) == 0.0  &&  floatx(var.imaginary) == 0.0 )
		return (real > var.real);
	else
		Throw_Domain_error( "imaginary must be zero" );
}

Complex sqroot(const Complex &var) throw(error_obj)
{
	if( floatx(var.imaginary) != 0.0 )
		return power( var, Complex( 0.5 ) );

	if( floatx(var.real) >= 0.0 )
		return Complex( sqrtx(floatx(var.real)) );
	else
		return Complex( Real(0), sqrtx(-floatx(var.real)) );
}

bool complex_almost_equal(const Complex &left,const Complex &right)
{
	Complex l,r;

	r = right;
	l = left;

	if( floatx(l.real) <= 1e-10  &&  floatx(l.real) >= -1e-10 )
		l.real = Real(1e-10);
	if( floatx(l.imaginary) <= 1e-10  && floatx(l.imaginary) >= -1e-10 )
		l.imaginary = Real(1e-10);
	if( floatx(r.real) <= 1e-10  &&  floatx(r.real) >= -1e-10 )
		r.real = Real(1e-10);
	if( floatx(r.imaginary) <= 1e-10  &&  floatx(r.imaginary) >= -1e-10 )
		r.imaginary = Real(1e-10);

	if( (floatx(l.real) * floatx(r.real)) < 0.0 ) // not same sign
		return false;
	else if( (floatx(l.imaginary) * floatx(r.imaginary)) < 0.0 )
		return false;
	else if( fabsx(log10x(floatx(l.real)) - log10x(floatx(r.real))) > 1e-10 )
		return false;
	else if( fabsx(log10x(floatx(l.imaginary)) - log10x(floatx(r.imaginary))) > 1e-10 )
		return false;

	return true;
}

Complex power(const Complex &left,const Complex &right) throw(error_obj)
{
	Complex tmp;

	tmp.imaginary = Real(0);

	if( floatx(right.imaginary) != 0.0 )
	{
		// let's use the fact that   a^b = ( e^(ln a) )^b = e^(b*ln a)
		tmp = expc( right * logc(left) );
	}
	else
	{
		if( floatx(left.imaginary) != 0.0 ) // we need to use de Moivres formula
		{
de_moivres:
			Complex tmp_abs,tmp_arg;

			tmp_abs = abs(left);
			tmp_arg = arg(left);
			tmp = powx( floatx(tmp_abs.real), floatx(right.real) ) * 
			           (Complex( cosx(floatx(right.real)*floatx(tmp_arg.real)),
			           sinx(floatx(right.real)*floatx(tmp_arg.real)) ) );
			// tmp = r^n * ( cos(n*phi) + i*sin(n*phi) )
		}
		else
		{
			if( left.real.isExact()  &&  right.real.isInteger() )
			{
				Integer nr;
				right.real.get_top( &nr );
				tmp.real = power( left.real, nr );
			}
			else if( floatx(left.real) < 0.0 )
			{
				if( right.real.isInteger() ) // checks if right.real == N
					tmp.real = Real( powx(floatx(left.real), floatx(right.real)) );
				else if((1.0/floatx(right.real)) == (truncx(1.0/floatx(right.real)))) // checks if right.real == 1/N
				{
					if(((1.0/floatx(right.real))/4.0) == (truncx((1.0/floatx(right.real))/4.0))) // N is dividable with 4
						goto de_moivres;
					else if(((1.0/floatx(right.real))/2.0) == (truncx((1.0/floatx(right.real))/2.0))) // N is dividable with 2
					{
						tmp.real = Real(0);
						tmp.imaginary = Real( powx(-floatx(left.real),floatx(right.real)) );
					}
					else
						tmp.real = Real( -powx(-floatx(left.real),floatx(right.real)) );
				}
				else
					goto de_moivres;
			}
			else
				tmp.real = Real( powx(floatx(left.real),floatx(right.real)) );
		}
	}
	return tmp;
}

Complex abs(const Complex &var) throw(error_obj)
{
	Real tmp(0);

	if( var.imaginary == tmp )
	{
		if( var.real < tmp )
			return Complex( -var.real );
		return var;
	}
	else
		return Complex( hypotx( floatx(var.real), floatx(var.imaginary) ) );
}


Complex expc(const Complex &val)
{
	Complex tmp;

	tmp.real = expx( floatx(val.real) );
	tmp.imaginary = tmp.real * Real( sinx( floatx(val.imaginary) ) );
	tmp.real = tmp.real * Real( cosx( floatx(val.imaginary) ) );

	return tmp;
}

Complex logc(const Complex &val)
{
//ln(z) = ln( abs(z) ) + i*arg(z)
	return Complex( logx( floatx( abs( val ).real ) ), arg( val ).real );
}

Complex log2c(const Complex &val)
{
//log2(z) = ln(z)/ln(2)
	return logc( val ) / Complex( logx( 2.0 ) );
}

Complex log10c(const Complex &val)
{
//log10(z) = ln(z)/ln(10)
	return logc( val ) / Complex( logx( 10.0 ) );
}

Complex sinc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( sinx( floatx(val.real) ) );
	else
	{
// sin(z) = ( e^(iz) - e^(-iz) ) / (2i)
		return ( expc(Complex( 0.0, 1.0 ) * val) - expc( Complex( 0.0, -1.0 ) * val ) ) /
		         Complex( 0.0, 2.0 );
	}
}

Complex cosc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( cosx( floatx(val.real) ) );
	else
	{
// cos(z) = ( e^(iz) + e^(-iz) ) / 2
		return ( expc(Complex( 0.0, 1.0 ) * val) + expc( Complex( 0.0, -1.0 ) * val ) ) /
		         Complex( 2.0 );
	}
}

Complex tanc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( tanx( floatx(val.real) ) );
	else
	{
// tan(z) = ( e^(2iz) -1 ) / ( i(e^(2iz) +1) )
		Complex tmp = expc( Complex( 0.0, 2.0 ) * val );
		return (tmp - Complex( 1.0 )) / ( Complex( 0.0, 1.0 ) * (tmp + Complex( 1.0 )) );
	}
}

Complex asinc(const Complex &val)
{
	floatx tmp = floatx(val.real);
	if( floatx(val.imaginary) == 0.0  &&  tmp <= 1.0  &&  tmp >= -1.0 )
		return Complex( cosx( tmp ) );
	else
	{
// asin(z) = -i*ln( iz + sqrt(1-z^2) )
		return Complex( 0.0, -1.0 ) * logc( Complex( 0.0, 1.0 ) * val +
		       sqroot( Complex( 1.0 ) - val*val ) );
	}
}

Complex acosc(const Complex &val)
{
	floatx tmp = floatx(val.real);
	if( floatx(val.imaginary) == 0.0  &&  tmp <= 1.0  &&  tmp >= -1.0 )
		return Complex( cosx( tmp ) );
	else
	{
// acos(z) = pi/2 + i*ln( iz + sqrt(1-z^2) )
		return Complex( PI/2.0 ) + Complex( 0.0, 1.0 ) * logc( Complex( 0.0, 1.0 ) * val + 
		       sqroot( Complex( 1.0 ) - val*val ) );
	}
}

Complex atanc(const Complex &val)
{
	if( floatx(val.imaginary) == 0.0 )
		return Complex( atanx( floatx(val.real) ) );
	else
	{
		// atan(z) = i/2 * ( ln(1-iz) - ln(1+iz) )
		return Complex( 0.0, 0.5 ) * ( logc( Complex( 1.0 ) - Complex( 0.0, 1.0 ) * val ) -
		         logc( Complex( 1.0 ) + Complex( 0.0, 1.0 ) * val ) );
	}
}

Complex sinhc(const Complex &val)
{
//sinh(z) = (exp(z) - exp(-z))/2
	return ( expc( val ) - expc( -val ) ) / Complex( 2.0 );
}

Complex coshc(const Complex &val)
{
//cosh(z) = (exp(z) + exp(-z))/2.
	return ( expc( val ) + expc( -val ) ) / Complex( 2.0 );
}

Complex tanhc(const Complex &val)
{
//tanh(z) = sinh(z)/cosh(z)
	return sinhc( val ) / coshc( val );  
}

Complex asinhc(const Complex &val)
{
//asinh(z) = ln(z + sqrt(z^2+1))
	return logc( val + sqroot( val*val + Complex(1.0) ) );
}

Complex acoshc(const Complex &val)
{
//acosh(z)=ln(z + sqrt(z^2-1))
	return logc( val + sqroot( val*val - Complex(1.0) ) );
}

Complex atanhc(const Complex &val)
{
//atanh(z)=ln( (1+z)/(1-z) ) / 2
	return logc( (Complex(1.0)+val)/(Complex(1.0)-val) ) / Complex(2.0);
}
